# Copyright (c) 2024 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Connects to the first FIDO device found which supports the PRF extension,
creates a new credential for it with the extension enabled, and uses it to
derive two separate secrets.
"""

import sys

from exampleutils import get_client

from fido2 import cbor
from fido2.cose import ESP256_SPLIT_ARKG_PLACEHOLDER, CoseKey
from fido2.ctap2.extensions import PreviewSignExtension
from fido2.server import Fido2Server
from fido2.utils import sha256, websafe_decode, websafe_encode

uv = "discouraged"

# Locate a suitable FIDO authenticator
client, info = get_client(
    lambda info: PreviewSignExtension.NAME in info.extensions,
    extensions=[PreviewSignExtension()],
)

server = Fido2Server({"id": "example.com", "name": "Example RP"}, attestation="none")
user = {"id": b"user_id", "name": "A. User"}

# Prepare parameters for makeCredential
create_options, state = server.register_begin(
    user,
    resident_key_requirement="discouraged",
    user_verification=uv,
    authenticator_attachment="cross-platform",
)

# Create a credential
result = client.make_credential(
    {
        **create_options["publicKey"],
        "extensions": {
            PreviewSignExtension.NAME: {
                "generateKey": {"algorithms": [ESP256_SPLIT_ARKG_PLACEHOLDER]}
            }
        },
    }
)

# Complete registration
auth_data = server.register_complete(state, result)
credentials = [auth_data.credential_data]
print("New credential created, with the sign extension.")

# PRF result:
sign_result = result.client_extension_results.previewSign
print("CREATE sign result", sign_result)
sign_key = sign_result.generated_key
if not sign_key:
    print(
        "Failed to create credential with sign extension",
        result.client_extension_results,
    )
    sys.exit(1)

# Extension output contains master public key
pk = CoseKey.parse(
    cbor.decode(websafe_decode(sign_key["publicKey"]))
)  # COSE key in bytes
print("public key", pk)

# Master public key contains blinding and KEM keys
# ARKG derive_public_key uses these
print("Blinding public key", pk.pk_bl)
print("KEM public key", pk.pk_kem)

# Arbitrary bytestring used for ctx, ikm
ctx = b"my-ctx-here"
ikm = b"my-ikm-here"
# Derived public key to verify with, and kh to send to Authenticator
pk2, args = pk.derive_public_key(ikm, ctx)
print("Derived public key", pk2)
print("COSE_Sign_Args for derived key", args)

# Prepare a message to sign
message = b"New message"
ph_data = sha256(message)

# Prepare parameters for getAssertion
request_options, state = server.authenticate_begin(credentials, user_verification=uv)


# Authenticate the credential
result = client.get_assertion(
    {
        **request_options["publicKey"],
        # Add extension outputs. We have only 1 credential in allowCredentials
        "extensions": {
            PreviewSignExtension.NAME: {
                "signByCredential": {
                    websafe_encode(credentials[0].credential_id): {
                        "keyHandle": sign_key.key_handle,
                        "tbs": ph_data,
                        "additionalArgs": cbor.encode(args),
                    },
                },
            }
        },
    }
)

# Only one cred in allowCredentials, only one response.
result = result.get_response(0)

sign_result = result.client_extension_results[PreviewSignExtension.NAME]
print("GET sign result", sign_result)

# Response contains a signature over message
signature = sign_result.get("signature")

print("Test verify signature", signature)
pk2.verify(message, websafe_decode(signature))
print("Signature verified with derived public key!")



import logging
from dataclasses import replace
from threading import Event, Thread

import pytest

from fido2.client import ClientError, DefaultClientDataCollector, Fido2Client
from fido2.cose import CoseKey
from fido2.ctap2 import Ctap2
from fido2.ctap2.credman import CredentialManagement
from fido2.ctap2.pin import ClientPin, PinProtocolV1, PinProtocolV2
from fido2.hid import (
    CAPABILITY,
    CtapHidDevice,
    list_descriptors,
    open_connection,
    open_device,
)

from . import TEST_PIN, CliInteraction, Printer

logger = logging.getLogger(__name__)


class DeviceManager:
    def __init__(self, printer, reader_name, ccid):
        self.printer = printer

        self.printer.print(
            "⚠️  Tests will now run against a connected FIDO authenticator. ⚠️ ",
            "",
            "You may be prompted to interact with the authenticator throughout these "
            "tests.",
            "",
            "          ☠️  WARNING! THESE TESTS ARE DESTRUCTIVE! ☠️ ",
            "ANY CREDENTIALS ON THIS AUTHENTICATOR WILL BE PERMANENTLY DELETED!",
        )

        self._transport = "usb"
        if reader_name:
            try:
                from fido2.pcsc import _list_readers
            except ImportError:
                pytest.exit("pyscard not installed, install package with 'pcsc' extra")

            readers = [
                r for r in _list_readers() if reader_name.lower() in r.name.lower()
            ]
            if len(readers) == 1:
                self._reader = readers[0]
                self._dev = self._connect_pcsc(self._reader, not ccid)
                if not ccid:
                    self._transport = "nfc"
            else:
                pytest.exit(f"No/Multiple NFC readers found matching '{reader_name}'")
        else:
            self._reader = None
            self._dev = self._select()

        if self.has_ctap2():
            info = Ctap2(self.device).info
            if info.transports_for_reset:
                self._can_reset = self._transport in info.transports_for_reset
            else:
                self._can_reset = True
            if info.options.get("clientPin") or info.options.get("uv"):
                self.printer.print(
                    "As a precaution, these tests will not run on an authenticator "
                    "which is configured with any form of UV. Factory reset the "
                    "authenticator prior to running tests against it."
                )
                pytest.exit("Authenticator must be in a newly-reset state!")

            if not self._can_reset:
                self.printer.print(
                    "⚠️  FACTORY RESET NOT ENABLED FOR THIS TRANSPORT! ⚠️ ",
                    "Some tests will be skipped, and the authenticator will be left "
                    "in a state where it will need to be factory reset through another"
                    "transport to be used.",
                )

        self.setup()

    def _select(self):
        event = Event()
        selected = []

        def select(descriptor):
            dev = CtapHidDevice(descriptor, open_connection(descriptor))
            # This client is only used for selection
            client = Fido2Client(
                dev,
                client_data_collector=DefaultClientDataCollector("https://example.com"),
            )
            try:
                while not event.is_set():
                    try:
                        client.selection(event)
                        selected.append(dev)
                        event.set()
                        break
                    except ClientError as e:
                        if e.code != ClientError.ERR.TIMEOUT:
                            raise
            except Exception:
                logger.debug(
                    f"Authenticator selection failed for {descriptor.path}",
                    exc_info=True,
                )

        self.printer.touch()

        descriptors: set[str | bytes] = set()
        threads: list[Thread] = []
        try:
            while not event.wait(0.1):
                if event.is_set():
                    break
                removed = set(descriptors)
                for d in list_descriptors():
                    if d.path not in descriptors:
                        descriptors.add(d.path)
                        t = Thread(target=select, args=(d,))
                        threads.append(t)
                        t.start()
                    else:
                        removed.remove(d.path)
                descriptors -= removed
        finally:
            event.set()

        for t in threads:
            # wait for child threads to finish
            t.join()

        if selected:
            logger.debug("Authenticator selected")
            return selected[0]
        else:
            pytest.exit("No Authenticator selected")

    def _connect_pcsc(self, reader, reconnect=True):
        from fido2.pcsc import CtapPcscDevice
        from smartcard.Exceptions import NoCardException, CardConnectionException
        from smartcard.ExclusiveConnectCardConnection import (
            ExclusiveConnectCardConnection,
        )

        logger.debug(f"(Re-)connect over NFC using reader: {reader.name}")
        event = Event()

        def _connect():
            connection = ExclusiveConnectCardConnection(reader.createConnection())
            return CtapPcscDevice(connection, reader.name)

        if not reconnect:
            return _connect()

        self.printer.remove(nfc=True)
        removed = False
        while not event.wait(0.5):
            try:
                dev = _connect()
                if removed:
                    dev.close()
                    event.wait(1.0)  # Wait for the device to settle
                    return _connect()
                dev.close()
            except CardConnectionException:
                pass  # Expected, ignore
            except (NoCardException, KeyError):
                if not removed:
                    self.printer.insert(nfc=True)
                    removed = True

        raise Exception("Failed to (re-)connect to Authenticator")

    @property
    def device(self):
        return self._dev

    def has_ctap2(self):
        return self.device.capabilities & CAPABILITY.CBOR

    @property
    def ctap2(self):
        if self.has_ctap2():
            return Ctap2(self.device)
        pytest.skip("Authenticator does not support CTAP 2")

    @property
    def info(self):
        return self.ctap2.get_info()

    @property
    def on_keepalive(self):
        prompted = [0]

        def on_keepalive(status):
            if status != prompted[0]:
                prompted[0] = status
                if status == 2:
                    self.printer.touch()

        return on_keepalive

    @property
    def client(self):
        return Fido2Client(
            self.device,
            client_data_collector=DefaultClientDataCollector("https://example.com"),
            user_interaction=CliInteraction(self.printer),
        )

    def _reconnect_usb(self):
        event = Event()
        dev_path = self._dev.descriptor.path
        info = Ctap2(self._dev).info
        logger.debug(f"Reconnect over USB: {dev_path}")

        self.printer.remove()
        ds = {d.path for d in list_descriptors()}
        while dev_path in ds:
            event.wait(0.5)
            ds = {d.path for d in list_descriptors()}
        self.printer.insert()
        ds2 = ds
        while True:
            event.wait(0.5)
            ds2 = {d.path for d in list_descriptors()}
            added = ds2 - ds
            if len(added) == 1:
                device = open_device(added.pop())
                info2 = Ctap2(device).info
                assert replace(
                    info, enc_identifier=None, enc_cred_store_state=None
                ) == replace(info2, enc_identifier=None, enc_cred_store_state=None)
                return device
            elif len(added) > 1:
                raise ValueError("Multiple Authenticators inserted")

    def _reconnect_ccid(self):
        name = self._reader.name
        info = Ctap2(self._dev).info
        self._dev.close(release=True)

        logger.debug(f"Reconnect over CCID: {name}")
        from fido2.pcsc import _list_readers

        self.printer.remove()
        event = Event()
        removed = False
        while not event.wait(0.5):
            readers = [r for r in _list_readers() if name == r.name]
            if removed and len(readers) == 1:
                self._reader = readers[0]
                dev = self._connect_pcsc(self._reader, False)
                info2 = Ctap2(dev).info
                assert replace(
                    info, enc_identifier=None, enc_cred_store_state=None
                ) == replace(info2, enc_identifier=None, enc_cred_store_state=None)

                return dev
            elif not removed and len(readers) == 0:
                self.printer.insert()
                removed = True

    def reconnect(self):
        if self._transport == "nfc":
            self._dev.close()
            self._dev.connect()
        elif self._reader:
            self._dev = self._reconnect_ccid()
        else:
            self._dev = self._reconnect_usb()
        return self._dev

    def _factory_reset(self, setup=False):
        if not self._can_reset:
            self.printer.print(
                "☠️  FACTORY RESET CALLED! ☠️ ",
                "",
                "This test should have been marked as requiring reset!",
            )
            pytest.exit("FACTORY RESET CALLED")

        self.printer.print("⚠️  PERFORMING FACTORY RESET! ⚠️ ")

        self.reconnect()

        if self.info.long_touch_for_reset:
            prompted = [0]

            def on_keepalive_reset(status):
                if status != prompted[0]:
                    prompted[0] = status
                    if status == 2:
                        self.printer.print(
                            "👉👉👉 Press and hold the Authenticator button..."
                        )
                    elif status == 1:
                        self.printer.print("✅ You can now release the button!")

        else:
            on_keepalive_reset = self.on_keepalive

        self.ctap2.reset(on_keepalive=on_keepalive_reset)

        if setup:
            self.setup()

    def setup(self):
        if self.has_ctap2() and ClientPin.is_supported(self.info):
            ClientPin(self.ctap2).set_pin(TEST_PIN)


@pytest.fixture(scope="session")
def printer(request):
    capmanager = request.config.pluginmanager.getplugin("capturemanager")
    return Printer(capmanager)


@pytest.fixture(scope="session", autouse=True)
def dev_manager(pytestconfig, printer):
    if pytestconfig.getoption("no_device"):
        pytest.skip("Skip device tests")

    reader = pytestconfig.getoption("reader")
    ccid = pytestconfig.getoption("ccid")
    manager = DeviceManager(printer, reader, ccid)

    yield manager

    # after the test, reset the device, if possible
    if manager._can_reset:
        manager._factory_reset()


@pytest.fixture(scope="session")
def factory_reset(dev_manager):
    if dev_manager._can_reset:
        return dev_manager._factory_reset
    pytest.skip("Requires factory reset")


@pytest.fixture
def device(dev_manager):
    return dev_manager.device


@pytest.fixture
def ctap2(dev_manager):
    return dev_manager.ctap2


@pytest.fixture
def on_keepalive(dev_manager):
    return dev_manager.on_keepalive


@pytest.fixture
def client(dev_manager):
    return dev_manager.client


@pytest.fixture
def info(ctap2):
    return ctap2.get_info()


@pytest.fixture(params=[CoseKey.for_alg(alg) for alg in CoseKey.supported_algorithms()])
def algorithm(request, info):
    alg_cls = request.param
    alg = {"alg": alg_cls.ALGORITHM, "type": "public-key"}
    if alg not in info.algorithms:
        pytest.skip(f"Algorithm {alg_cls.__name__} not supported")
    return alg


@pytest.fixture(params=[PinProtocolV1, PinProtocolV2])
def pin_protocol(request, info):
    proto = request.param
    if proto.VERSION not in info.pin_uv_protocols:
        pytest.skip(f"PIN/UV protocol {proto.VERSION} not supported")

    all_protocols = ClientPin.PROTOCOLS
    # Ensure we always negotiate only the selected protocol
    ClientPin.PROTOCOLS = [proto]
    yield proto()

    ClientPin.PROTOCOLS = all_protocols


@pytest.fixture
def clear_creds(dev_manager):
    # Clears any discoverable credentials after a test
    yield None
    clientpin = ClientPin(dev_manager.ctap2)
    token = clientpin.get_pin_token(TEST_PIN, ClientPin.PERMISSION.CREDENTIAL_MGMT)
    credman = CredentialManagement(dev_manager.ctap2, clientpin.protocol, token)
    for rp in credman.enumerate_rps():
        for cred in credman.enumerate_creds(rp[4]):
            credman.delete_cred(cred[7])
    assert len(credman.enumerate_rps()) == 0



import os

import pytest
from fido2 import cbor
from fido2.client import DefaultClientDataCollector, Fido2Client
from fido2.cose import ESP256_SPLIT_ARKG_PLACEHOLDER, CoseKey
from fido2.ctap2.extensions import PreviewSignExtension
from fido2.server import Fido2Server
from fido2.utils import sha256, websafe_decode, websafe_encode

from . import CliInteraction


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if PreviewSignExtension.NAME not in dev_manager.info.extensions:
        pytest.skip("previewSign not supported by authenticator")


def test_client_arkg_p256(device, printer):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp, attestation="none")
    user = {"id": b"user_id", "name": "A. User"}
    uv = "discouraged"

    client = Fido2Client(
        device,
        client_data_collector=DefaultClientDataCollector("https://example.com"),
        user_interaction=CliInteraction(printer),
        extensions=[PreviewSignExtension()],
    )

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="discouraged",
        user_verification=uv,
        authenticator_attachment="cross-platform",
    )

    # Create a credential with the sign extension
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {
                PreviewSignExtension.NAME: {
                    "generateKey": {
                        "algorithms": [ESP256_SPLIT_ARKG_PLACEHOLDER],
                    }
                }
            },
        }
    )

    auth_data = server.register_complete(state, result)
    credentials = [auth_data.credential_data]

    sign_result = result.client_extension_results.previewSign
    sign_key = sign_result.generated_key
    assert sign_key is not None

    # Parse the master public key
    pk = CoseKey.parse(cbor.decode(websafe_decode(sign_key["publicKey"])))

    # Derive a public key using ARKG
    ctx = b"python-fido2.test_sign_extension_v4"
    ikm = os.urandom(32)
    pk2, args = pk.derive_public_key(ikm, ctx)

    # Prepare a message to sign
    message = b"test message"
    ph_data = sha256(message)

    # Prepare parameters for getAssertion
    request_options, state = server.authenticate_begin(
        credentials, user_verification=uv
    )

    # Authenticate with the sign extension
    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {
                PreviewSignExtension.NAME: {
                    "signByCredential": {
                        websafe_encode(credentials[0].credential_id): {
                            "keyHandle": sign_key.key_handle,
                            "tbs": ph_data,
                            "additionalArgs": cbor.encode(args),
                        },
                    },
                }
            },
        }
    )

    response = result.get_response(0)
    sign_result = response.client_extension_results[PreviewSignExtension.NAME]
    signature = sign_result.get("signature")
    assert signature is not None

    # Verify the signature with the derived public key
    pk2.verify(message, websafe_decode(signature))



from typing import Any

import pytest
from fido2.ctap import CtapError
from fido2.ctap2.credman import CredentialManagement
from fido2.ctap2.pin import ClientPin
from fido2.server import Fido2Server

from . import TEST_PIN


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if not CredentialManagement.is_supported(dev_manager.info):
        pytest.skip("CredentialManagement not supported by authenticator")


def get_credman(ctap2, pin_protocol, permissions=ClientPin.PERMISSION.CREDENTIAL_MGMT):
    token = ClientPin(ctap2, pin_protocol).get_pin_token(TEST_PIN, permissions)
    return CredentialManagement(ctap2, pin_protocol, token)


def test_list_and_delete(client, ctap2, pin_protocol, algorithm):
    # Ensure no credentials exist initially
    credman = get_credman(ctap2, pin_protocol)
    metadata = credman.get_metadata()
    assert metadata[CredentialManagement.RESULT.EXISTING_CRED_COUNT] == 0
    remaining = metadata[CredentialManagement.RESULT.MAX_REMAINING_COUNT]
    assert remaining > 0

    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
    )

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "pubKeyCredParams": [algorithm],
            "extensions": {"credProps": True},
        }
    )

    # Need new PIN token as old one is expired by make_credential
    credman = get_credman(ctap2, pin_protocol)

    metadata = credman.get_metadata()
    assert metadata[CredentialManagement.RESULT.EXISTING_CRED_COUNT] == 1
    assert metadata[CredentialManagement.RESULT.MAX_REMAINING_COUNT] < remaining

    # Complete registration
    auth_data = server.register_complete(state, result)
    cred = auth_data.credential_data
    assert cred.public_key[3] == algorithm["alg"]

    rps = credman.enumerate_rps()
    assert len(rps) == 1

    # Not all keys are required in response, but those that are should match
    for k, v in rps[0][CredentialManagement.RESULT.RP].items():
        assert rp[k] == v

    rp_id_hash = rps[0][CredentialManagement.RESULT.RP_ID_HASH]
    creds = credman.enumerate_creds(rp_id_hash)
    assert len(creds) == 1
    assert creds[0][CredentialManagement.RESULT.USER] == user
    assert (
        creds[0][CredentialManagement.RESULT.CREDENTIAL_ID]["id"] == cred.credential_id
    )
    assert creds[0][CredentialManagement.RESULT.PUBLIC_KEY] == cred.public_key
    assert not creds[0].get(CredentialManagement.RESULT.THIRD_PARTY_PAYMENT)

    credman.delete_cred(creds[0][CredentialManagement.RESULT.CREDENTIAL_ID])
    metadata = credman.get_metadata()
    assert metadata[CredentialManagement.RESULT.EXISTING_CRED_COUNT] == 0
    assert metadata[CredentialManagement.RESULT.MAX_REMAINING_COUNT] == remaining


def test_update(client, ctap2, pin_protocol):
    if not CredentialManagement.is_update_supported(ctap2.info):
        pytest.skip("ClientPin update not supported")

    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User", "displayName": "Display Name"}

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
    )

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"credProps": True},
        }
    )
    auth_data = server.register_complete(state, result)
    cred_id = {"id": auth_data.credential_data.credential_id, "type": "public-key"}

    credman = get_credman(ctap2, pin_protocol)
    rps = credman.enumerate_rps()
    rp_id_hash = rps[0][CredentialManagement.RESULT.RP_ID_HASH]

    # Check user data
    creds = credman.enumerate_creds(rp_id_hash)
    assert len(creds) == 1

    # Authenticators may or may not store name/displayName
    stores_name = "name" in creds[0][CredentialManagement.RESULT.USER]
    stores_display_name = "displayName" in creds[0][CredentialManagement.RESULT.USER]

    if not stores_name:
        del user["name"]
    if not stores_display_name:
        del user["displayName"]

    assert creds[0][CredentialManagement.RESULT.USER] == user

    # Update user data
    user2: dict[str, Any] = {"id": b"user_id"}
    if stores_name:
        user2["name"] = "A. User 2"
    if stores_display_name:
        user2["displayName"] = "Display Name 2"

    credman.update_user_info(cred_id, user2)

    creds = credman.enumerate_creds(rp_id_hash)
    assert len(creds) == 1
    assert creds[0][CredentialManagement.RESULT.USER] == user2
    assert creds[0][CredentialManagement.RESULT.CREDENTIAL_ID] == cred_id
    assert not creds[0].get(CredentialManagement.RESULT.THIRD_PARTY_PAYMENT)

    # Test deleting fields
    user3 = {"id": b"user_id"}
    credman.update_user_info(cred_id, user3)
    creds = credman.enumerate_creds(rp_id_hash)
    assert len(creds) == 1
    assert creds[0][CredentialManagement.RESULT.USER] == user3
    assert creds[0][CredentialManagement.RESULT.CREDENTIAL_ID] == cred_id
    assert not creds[0].get(CredentialManagement.RESULT.THIRD_PARTY_PAYMENT)

    # Clean up
    credman.delete_cred(cred_id)


def test_missing_permissions(ctap2, pin_protocol):
    if not ClientPin.is_token_supported(ctap2.info):
        pytest.skip("Permissions not supported")

    credman = get_credman(ctap2, pin_protocol, ClientPin.PERMISSION.LARGE_BLOB_WRITE)

    with pytest.raises(CtapError, match="PIN_AUTH_INVALID"):
        credman.get_metadata()


def test_read_only_management(dev_manager, pin_protocol):
    if not CredentialManagement.is_readonly_supported(dev_manager.info):
        pytest.skip("Persistent PUAT not supported")

    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
    )

    token = ClientPin(dev_manager.ctap2, pin_protocol).get_pin_token(
        TEST_PIN, ClientPin.PERMISSION.PERSISTENT_CREDENTIAL_MGMT
    )

    # Get cred_store_state, enc_identifier before reconnect
    cred_state = dev_manager.ctap2.get_info().get_cred_store_state(token)
    ident = dev_manager.ctap2.get_info().get_identifier(token)

    # Create a credential
    result = dev_manager.client.make_credential(create_options["publicKey"])
    auth_data = server.register_complete(state, result)
    cred_id = {"id": auth_data.credential_data.credential_id, "type": "public-key"}
    rp_id_hash = server.rp.id_hash

    # Verify cred_store_state has changed
    if cred_state:
        new_cred_state = dev_manager.ctap2.get_info().get_cred_store_state(token)
        assert new_cred_state != cred_state
        cred_state = new_cred_state

    # Test token before and after reconnect
    for reconnect in (False, True):
        if reconnect:
            dev_manager.reconnect()

        # Use persistent token
        credman = CredentialManagement(dev_manager.ctap2, pin_protocol, token)

        # Test metadata
        assert credman.get_metadata()[1] == 1

        # Test enumerate RPs and creds
        rps = credman.enumerate_rps()
        assert len(rps) == 1
        creds = credman.enumerate_creds(rp_id_hash)
        assert len(creds) == 1

        # Ensure update isn't allowed
        user2 = {"id": b"user_id", "name": "A. User 2"}
        with pytest.raises(CtapError, match="PIN_AUTH_INVALID"):
            credman.update_user_info(cred_id, user2)

        # Ensure delete isn't allowed
        with pytest.raises(CtapError, match="PIN_AUTH_INVALID"):
            credman.delete_cred(cred_id)

        # Ensure cred_store_state is the same
        assert dev_manager.ctap2.get_info().get_cred_store_state(token) == cred_state

    # Compare enc_identifier after reconnect
    assert dev_manager.ctap2.get_info().get_identifier(token) == ident

    # Use new (non-persistent) PIN token
    credman = get_credman(dev_manager.ctap2, pin_protocol)
    credman.delete_cred(cred_id)

    # Verify cred_store_state has changed
    if cred_state:
        new_cred_state = dev_manager.ctap2.get_info().get_cred_store_state(token)
        assert new_cred_state != cred_state
        cred_state = new_cred_state



# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Connects to each FIDO device found, and causes them all to blink until the user
triggers one to select it. A new credential is created for that authenticator,
and the operation is cancelled for the others.
"""

import sys
from getpass import getpass
from threading import Event, Thread

from fido2.client import (
    ClientError,
    DefaultClientDataCollector,
    Fido2Client,
    UserInteraction,
)
from fido2.hid import CtapHidDevice

# Locate a device
devs = list(CtapHidDevice.list_devices())
if not devs:
    print("No FIDO device found")
    sys.exit(1)


# Handle user interaction
class CliInteraction(UserInteraction):
    def prompt_up(self):
        print("\nTouch your authenticator device now...\n")

    def request_pin(self, permissions, rd_id):
        return getpass("Enter PIN: ")

    def request_uv(self, permissions, rd_id):
        print("User Verification required.")
        return True


cli_interaction = CliInteraction()
clients = [
    Fido2Client(
        d,
        client_data_collector=DefaultClientDataCollector("https://example.com"),
        user_interaction=cli_interaction,
    )
    for d in devs
]

# Prepare parameters for makeCredential
rp = {"id": "example.com", "name": "Example RP"}
user = {"id": b"user_id", "name": "A. User"}
challenge = b"Y2hhbGxlbmdl"
cancel = Event()
selected = None


def select(client):
    global selected
    try:
        client.selection(cancel)
        selected = client
    except ClientError as e:
        if e.code != ClientError.ERR.TIMEOUT:
            raise
        else:
            return
    cancel.set()


print("\nTouch the authenticator you wish to use...\n")

threads = []
for client in clients:
    t = Thread(target=select, args=(client,))
    threads.append(t)
    t.start()

for t in threads:
    t.join()

if cancel.is_set():
    print("Authenticator selected, making credential...")

    result = selected.make_credential(
        {
            "rp": rp,
            "user": user,
            "challenge": challenge,
            "pubKeyCredParams": [{"type": "public-key", "alg": -7}],
        },
    )

    print("New credential created!")
    response = result.response
    print("ATTESTATION OBJECT:", response.attestation_object)
    print()
    print("CREDENTIAL DATA:", response.attestation_object.auth_data.credential_data)
else:
    print("Operation timed out!")



import os

import pytest
from fido2.server import Fido2Server


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if "credBlob" not in dev_manager.info.extensions:
        pytest.skip("CredBlob not supported by authenticator")


def test_read_write(client, ctap2, clear_creds):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
        user_verification="required",
    )

    # Create a credential
    blob = os.urandom(32)
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"credBlob": blob},
        }
    )
    auth_data = server.register_complete(state, result)
    credentials = [auth_data.credential_data]

    assert auth_data.extensions
    assert auth_data.extensions["credBlob"] is True

    request_options, state = server.authenticate_begin(
        credentials, user_verification="required"
    )

    selection = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {"getCredBlob": True},
        }
    )
    result = selection.get_response(0)

    assert result.response.authenticator_data.extensions.get("credBlob") == blob



def pytest_addoption(parser):
    parser.addoption("--reader", action="store")
    parser.addoption("--no-device", action="store_true")
    parser.addoption("--ep-rp-id", action="store")
    parser.addoption("--ccid", action="store_true")



# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Connects to the first FIDO device found (starts from USB, then looks into NFC),
creates a new credential for it, and authenticates the credential.
This works with both FIDO 2.0 devices as well as with U2F devices.
On Windows, the native WebAuthn API will be used.
"""

from exampleutils import get_client

from fido2.server import Fido2Server

# Locate a suitable FIDO authenticator
client, info = get_client(lambda info: info.options.get("rk"))

# Prefer UV if supported and configured
uv = "discouraged"
if info and info.options.get("uv") or info.options.get("bioEnroll"):
    uv = "preferred"
    print("Authenticator is configured for User Verification")

server = Fido2Server({"id": "example.com", "name": "Example RP"}, attestation="direct")

user = {"id": b"user_id", "name": "A. User"}

# Prepare parameters for makeCredential
create_options, state = server.register_begin(
    user,
    resident_key_requirement="required",
    user_verification=uv,
    authenticator_attachment="cross-platform",
)

# Create a credential
result = client.make_credential(
    {
        **create_options["publicKey"],
        # This extension isn't needed, but can be used to verify that the created
        # credential uses resident key
        "extensions": {"credProps": True},
    }
)


# Complete registration
auth_data = server.register_complete(state, result)
credentials = [auth_data.credential_data]

print("New credential created!")
response = result.response

print("CLIENT DATA:", response.client_data)
print("ATTESTATION OBJECT:", response.attestation_object)
print()
print("CREDENTIAL DATA:", auth_data.credential_data)
print()
print("Credential Properties:", result.client_extension_results.get("credProps"))

# credProps:
cred_props = result.client_extension_results.get("credProps")
print("CredProps", cred_props)


# Prepare parameters for getAssertion
request_options, state = server.authenticate_begin(user_verification=uv)

# Authenticate the credential
selection = client.get_assertion(request_options["publicKey"])
result = selection.get_response(0)  # There may be multiple responses, get the first.


# Complete authenticator
server.authenticate_complete(state, credentials, result)

print("Credential authenticated!")
response = result.response

print("USER ID:", response.user_handle)
print("CLIENT DATA:", response.client_data)
print()
print("AUTHENTICATOR DATA:", response.authenticator_data)



import unittest

from fido2.server import Fido2Server
from fido2.utils import websafe_encode
from fido2.webauthn import (
    AttestedCredentialData,
    AuthenticationResponse,
    AuthenticatorAssertionResponse,
    AuthenticatorData,
    CollectedClientData,
    PublicKeyCredentialRpEntity,
    UserVerificationRequirement,
)

from .test_ctap2 import _ATT_CRED_DATA, _CRED_ID


class TestPublicKeyCredentialRpEntity(unittest.TestCase):
    def test_id_hash(self):
        rp = PublicKeyCredentialRpEntity(name="Example", id="example.com")
        rp_id_hash = (
            b"\xa3y\xa6\xf6\xee\xaf\xb9\xa5^7\x8c\x11\x804\xe2u\x1eh/"
            b"\xab\x9f-0\xab\x13\xd2\x12U\x86\xce\x19G"
        )
        self.assertEqual(rp.id_hash, rp_id_hash)


USER = {"id": b"user_id", "name": "A. User"}


class TestFido2Server(unittest.TestCase):
    def test_register_begin_rp(self):
        rp = PublicKeyCredentialRpEntity(name="Example", id="example.com")
        server = Fido2Server(rp)

        request, state = server.register_begin(USER)

        self.assertEqual(
            request["publicKey"]["rp"], {"id": "example.com", "name": "Example"}
        )

    def test_register_begin_custom_challenge(self):
        rp = PublicKeyCredentialRpEntity(name="Example", id="example.com")
        server = Fido2Server(rp)

        challenge = b"1234567890123456"
        request, state = server.register_begin(USER, challenge=challenge)

        self.assertEqual(request["publicKey"]["challenge"], websafe_encode(challenge))

    def test_register_begin_custom_challenge_too_short(self):
        rp = PublicKeyCredentialRpEntity(name="Example", id="example.com")
        server = Fido2Server(rp)

        challenge = b"123456789012345"
        with self.assertRaises(ValueError):
            request, state = server.register_begin(USER, challenge=challenge)

    def test_authenticate_complete_invalid_signature(self):
        rp = PublicKeyCredentialRpEntity(name="Example", id="example.com")
        server = Fido2Server(rp)

        state = {
            "challenge": "GAZPACHO!",
            "user_verification": UserVerificationRequirement.PREFERRED,
        }
        client_data = CollectedClientData.create(
            CollectedClientData.TYPE.GET,
            "GAZPACHO!",
            "https://example.com",
        )
        _AUTH_DATA = bytes.fromhex(
            "A379A6F6EEAFB9A55E378C118034E2751E682FAB9F2D30AB13D2125586CE1947010000001D"
        )
        response = AuthenticationResponse(
            raw_id=_CRED_ID,
            response=AuthenticatorAssertionResponse(
                client_data=client_data,
                authenticator_data=AuthenticatorData(_AUTH_DATA),
                signature=b"INVALID",
            ),
        )

        with self.assertRaisesRegex(ValueError, "Invalid signature."):
            server.authenticate_complete(
                state,
                [AttestedCredentialData(_ATT_CRED_DATA)],
                response,
            )



import pytest
from fido2.client import Fido2Client
from fido2.ctap2.credman import CredentialManagement
from fido2.ctap2.extensions import (
    PaymentCredentialInstrument,
    PaymentCurrencyAmount,
    ThirdPartyPaymentExtension,
)
from fido2.ctap2.pin import ClientPin
from fido2.payment import (
    CollectedClientAdditionalPaymentData,
    PaymentClientDataCollector,
)
from fido2.server import Fido2Server

from . import TEST_PIN, CliInteraction


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if "thirdPartyPayment" not in dev_manager.info.extensions:
        pytest.skip("thirdPartyPayment not supported by authenticator")


def test_payment_extension(device, printer, ctap2, pin_protocol):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    # Prepare parameters for makeCredential
    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
        user_verification="required",
        authenticator_attachment="cross-platform",
    )

    client = Fido2Client(
        device,
        client_data_collector=PaymentClientDataCollector("https://example.com"),
        user_interaction=CliInteraction(printer, TEST_PIN),
        extensions=[ThirdPartyPaymentExtension()],
    )

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"payment": {"isPayment": True}},
        }
    )

    # Complete registration
    auth_data = server.register_complete(state, result)
    credentials = [auth_data.credential_data]

    print("Payment credential created!")

    # Test flag in Credential Management
    token = ClientPin(ctap2, pin_protocol).get_pin_token(
        TEST_PIN, ClientPin.PERMISSION.CREDENTIAL_MGMT
    )
    cm = CredentialManagement(ctap2, pin_protocol, token)
    rps = cm.enumerate_rps()
    rp_id_hash = rps[0][4]
    creds = cm.enumerate_creds(rp_id_hash)
    assert creds[0][CredentialManagement.RESULT.THIRD_PARTY_PAYMENT] is True

    # Prepare parameters for getAssertion
    request_options, state = server.authenticate_begin(
        credentials, user_verification="required"
    )

    # Prepare payment options
    payment = CollectedClientAdditionalPaymentData(
        rp_id="example.com",
        top_origin="https://top.example.com",
        payee_name="Mr. Payee",
        payee_origin="https://payee.example.com",
        total=PaymentCurrencyAmount(
            currency="USD",
            value="1.00",
        ),
        instrument=PaymentCredentialInstrument(
            display_name="My Payment",
            icon="https://example.com/icon.png",
        ),
    )

    # Authenticate the credential
    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {
                "payment": dict(payment, isPayment=True),
            },
        }
    )

    # Only one cred in allowCredentials, only one response.
    result = result.get_response(0)

    # Verify that the key includes the payment extension
    assert result.response.authenticator_data.extensions
    assert result.response.authenticator_data.extensions["thirdPartyPayment"] is True

    # Verify that the client has added the payment data
    assert result.response.client_data.type == "payment.get"
    assert result.response.client_data.payment == payment



import os
from collections.abc import Callable
from enum import IntEnum
from typing import Mapping, cast

import cryptography.exceptions
import pytest
from fido2 import cbor
from fido2.cose import ESP256_SPLIT_ARKG_PLACEHOLDER, CoseKey
from fido2.ctap import CtapError
from fido2.ctap2.pin import ClientPin
from fido2.utils import sha256
from fido2.webauthn import AttestationObject

from . import TEST_PIN


class AuthenticatorInput(IntEnum):
    KH = 2
    ALG = 3
    FLAGS = 4
    TBS = 6
    ARGS = 7


class AuthenticatorOutput(IntEnum):
    ALG = 3
    FLAGS = 4
    SIG = 6
    ATT_OBJ = 7


EMPTY_JSON_HASH = sha256(b"{}")
RP = {"id": "example.com", "name": "Example RP"}
USER = {"id": b"user_id", "name": "A. User"}
"""Not truly _all_ algorithms, but all we have prototyped"""
ALL_ALGORITHMS = [
    -9,  # ESP256
    -7,  # ES256
    -300,  # Requested assignment for ESP256-split https://www.ietf.org/archive/id/draft-lundberg-cose-two-party-signing-algs-06.html#name-ecdsa
    -70009,  # Placeholder value for ESP256-split used by some prototypes
    ESP256_SPLIT_ARKG_PLACEHOLDER,  # Placeholder for ESP256-split-ARKG https://www.ietf.org/archive/id/draft-bradleylundberg-cfrg-arkg-10.html#name-cose-algorithms
]
PREHASH_ALGS = [
    ESP256_SPLIT_ARKG_PLACEHOLDER,
    -300,
    -70009,  # Placeholder value for ESP256-split used by some prototypes
]


# These tests use only the raw authenticator inputs/outputs so the
# python-fido2 library doesn't need to maintain backwards compatibility of
# the client layer of the extension.


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if "previewSign" not in dev_manager.info.extensions:
        pytest.skip("previewSign not supported by authenticator")


def generate_key_inputs(algorithms, flags=0b000):
    return {
        AuthenticatorInput.ALG: algorithms,
        AuthenticatorInput.FLAGS: flags,
    }


def parse_generate_key_outputs(response):
    exts = response.auth_data.extensions or {}
    if "previewSign" in exts:
        assert AuthenticatorOutput.ALG in exts["previewSign"]
        algorithm = exts["previewSign"][AuthenticatorOutput.ALG]
        assert "previewSign" in response.unsigned_extension_outputs
        unsigned_outputs = response.unsigned_extension_outputs["previewSign"]

        assert AuthenticatorOutput.ATT_OBJ in unsigned_outputs
        att_obj_cbor = cast(
            Mapping, cbor.decode(unsigned_outputs[AuthenticatorOutput.ATT_OBJ])
        )
        att_obj = AttestationObject.create(
            att_obj_cbor[1], att_obj_cbor[2], att_obj_cbor[3]
        )

        key_handle = att_obj.auth_data.credential_data.credential_id
        public_key = att_obj.auth_data.credential_data.public_key
        return algorithm, key_handle, public_key, att_obj

    else:
        return None


def sign_inputs(cred, tbs, additional_args=None):
    sign_inputs = {
        AuthenticatorInput.KH: cred.key_handle,
        AuthenticatorInput.TBS: sha256(tbs) if cred.algorithm in PREHASH_ALGS else tbs,
    }
    if additional_args is not None:
        sign_inputs[AuthenticatorInput.ARGS] = cbor.encode(additional_args)
    return sign_inputs


def parse_sign_outputs(response):
    exts = response.auth_data.extensions
    if "previewSign" in exts and AuthenticatorOutput.SIG in exts["previewSign"]:
        signature = exts["previewSign"][AuthenticatorOutput.SIG]
        return signature

    else:
        return None


class Credential:
    def __init__(
        self,
        response,
        flags: int,
        options,
        extensions,
        generated_key: tuple[int, bytes, CoseKey, AttestationObject],
    ):
        self.response = response
        self.flags = flags
        self.options = options
        self.extensions = extensions
        algorithm, key_handle, public_key, att_obj = generated_key or [None] * 4
        self.algorithm = algorithm
        self.key_handle = key_handle
        self.public_key = public_key
        self.att_obj = att_obj

    def __repr__(self):
        return repr(
            (
                self.response,
                self.flags,
                self.options,
                self.extensions,
                self.algorithm,
                self.key_handle,
                self.public_key,
                self.att_obj,
            )
        )


class CredentialCache:
    def __init__(self, generate_key):
        self.generate_key = generate_key
        self.cache = []

    def make_cred(
        self,
        cache_filter: Callable[[Credential], bool],
        algorithms: list[int],
        flags=0b000,
        options=None,
        extensions={},
    ) -> Credential | None:
        existing = next((cred for cred in self.cache if cache_filter(cred)), None)
        if existing is not None:
            return existing
        else:
            cred = self.generate_key(algorithms, flags, options, extensions)
            self.cache.append(cred)
            return cred

    def make_cred_or_skip(self, cache_filter, algorithms, *args, **kwargs):
        try:
            return self.make_cred(cache_filter, algorithms, *args, **kwargs)
        except CtapError as e:
            if e.code == CtapError.ERR.UNSUPPORTED_ALGORITHM:
                pytest.skip(f"Algorithms {algorithms} not supported")
            raise e


@pytest.fixture(scope="module")
def credential_cache(generate_key):
    return CredentialCache(generate_key)


@pytest.fixture(scope="module")
def generate_key(dev_manager):
    def make_cred(
        algorithms: list[int],
        flags=0b000,
        options=None,
        extensions={},
    ) -> Credential:
        ext_inputs = generate_key_inputs(algorithms, flags)

        pin_protocol = None
        pin_uv_param = None
        if options is not None and options["rk"]:
            client_pin = ClientPin(dev_manager.ctap2)
            pin_protocol = client_pin.protocol.VERSION
            pin_token = client_pin.get_pin_token(TEST_PIN)
            pin_uv_param = client_pin.protocol.authenticate(pin_token, EMPTY_JSON_HASH)

        response = dev_manager.ctap2.make_credential(
            EMPTY_JSON_HASH,
            RP,
            USER,
            [{"type": "public-key", "alg": -7}],
            options=options,
            extensions={"previewSign": ext_inputs, **extensions},
            pin_uv_param=pin_uv_param,
            pin_uv_protocol=pin_protocol,
            on_keepalive=dev_manager.on_keepalive,
        )
        return Credential(
            response, flags, options, extensions, parse_generate_key_outputs(response)
        )

    return make_cred


@pytest.fixture(scope="module")
def sign(dev_manager):
    def get_assertion(
        cred: Credential,
        tbs,
        additional_args=None,
        up=False,
    ):
        ext_inputs = sign_inputs(cred, tbs, additional_args)
        response = dev_manager.ctap2.get_assertion(
            RP["id"],
            EMPTY_JSON_HASH,
            allow_list=[
                {
                    "type": "public-key",
                    "id": cred.response.auth_data.credential_data.credential_id,
                }
            ],
            options={"up": up},
            extensions={"previewSign": ext_inputs},
            on_keepalive=dev_manager.on_keepalive,
        )
        return (
            response,
            parse_sign_outputs(response),
        )

    return get_assertion


def if_arkg(algorithm, public_key):
    if algorithm == -65539:
        arkg_pub_seed = public_key
        assert arkg_pub_seed[3] == -65700, "Expected alg: ARKG-P256"
        assert arkg_pub_seed[-3] == -9, "Expected dkalg: ESP256"
        arkg_ikm = os.urandom(32)
        arkg_ctx = b"python-fido2.test_sign_extension_v4"
        return arkg_pub_seed.derive_public_key(arkg_ikm, arkg_ctx)
    else:
        return public_key, None


def test_esp256_split(credential_cache, sign):
    algorithms = [
        -300,  # Requested assignment for ESP256-split https://www.ietf.org/archive/id/draft-lundberg-cose-two-party-signing-algs-06.html#name-ecdsa
        -70009,  # Placeholder value used by some prototypes
    ]
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.algorithm in algorithms and cred.flags == 0b000,
        algorithms,
    )
    assert cred.algorithm in algorithms

    tbs = os.urandom(32)
    response, signature = sign(cred, tbs)

    assert signature is not None
    cred.public_key.verify(tbs, signature)


def test_esp256(credential_cache, sign):
    algorithms = [-9]
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.algorithm in algorithms and cred.flags == 0b000,
        algorithms,
    )
    assert cred.algorithm in algorithms

    tbs = os.urandom(32)
    response, signature = sign(cred, tbs)

    assert signature is not None
    cred.public_key.verify(tbs, signature)


def test_es256(credential_cache, sign):
    algorithms = [-7]
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.algorithm in algorithms and cred.flags == 0b000,
        algorithms,
    )
    assert cred.algorithm in algorithms

    tbs = os.urandom(32)
    response, signature = sign(cred, tbs)

    assert signature is not None
    cred.public_key.verify(tbs, signature)


def test_esp256_split_arkg(credential_cache, sign):
    algorithms = [
        -65539,  # Placeholder for ESP256-split-ARKG https://www.ietf.org/archive/id/draft-bradleylundberg-cfrg-arkg-10.html#name-cose-algorithms
    ]
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.algorithm in algorithms and cred.flags == 0b000,
        algorithms,
    )
    assert cred.algorithm in algorithms

    public_key, args = if_arkg(cred.algorithm, cred.public_key)

    tbs = os.urandom(32)
    response, signature = sign(cred, tbs, additional_args=args)

    assert signature is not None
    public_key.verify(tbs, signature)


def test_two_keys_same_alg(credential_cache, sign):
    algorithms = ALL_ALGORITHMS
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.algorithm in algorithms and cred.flags == 0b000,
        algorithms,
    )
    assert cred.algorithm in algorithms

    algorithms2 = [cred.algorithm]
    cred2 = credential_cache.make_cred_or_skip(
        lambda cred2: (
            cred2.algorithm == cred.algorithm
            and cred2.flags == 0b000
            and cred2.response.auth_data.credential_data.credential_id
            != cred.response.auth_data.credential_data.credential_id
        ),
        algorithms2,
    )
    assert cred2 is not None

    public_key, args = if_arkg(cred.algorithm, cred.public_key)

    tbs = os.urandom(32)
    response, signature = sign(cred, tbs, additional_args=args)

    assert signature is not None
    public_key.verify(tbs, signature)

    public_key2, _ = if_arkg(cred2.algorithm, cred2.public_key)

    assert cred2.algorithm in algorithms2
    assert cred2.algorithm == cred.algorithm
    assert cred2.key_handle != cred.key_handle
    assert cred2.public_key != cred.public_key
    assert cred2.att_obj != cred.att_obj

    with pytest.raises(cryptography.exceptions.InvalidSignature):
        public_key2.verify(tbs, signature)


def test_two_keys_different_alg(credential_cache, sign):
    algorithms = ALL_ALGORITHMS
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.algorithm in algorithms and cred.flags == 0b000,
        algorithms,
    )
    assert cred.algorithm in algorithms

    algorithms2 = [alg for alg in algorithms if alg != cred.algorithm]
    cred2 = credential_cache.make_cred_or_skip(
        lambda cred2: (
            cred2.algorithm in algorithms2
            and cred2.flags == 0b000
            and cred2.response.auth_data.credential_data.credential_id
            != cred.response.auth_data.credential_data.credential_id
            and cred2.algorithm != cred.algorithm
        ),
        algorithms2,
    )
    assert cred2 is not None

    public_key, args = if_arkg(cred.algorithm, cred.public_key)

    tbs = os.urandom(32)
    response, signature = sign(cred, tbs, additional_args=args)

    assert signature is not None
    public_key.verify(tbs, signature)

    public_key2, _ = if_arkg(cred2.algorithm, cred2.public_key)

    assert cred2.algorithm in algorithms2
    assert cred2.key_handle != cred.key_handle
    assert cred2.public_key != cred.public_key
    assert cred2.att_obj != cred.att_obj

    with pytest.raises(cryptography.exceptions.InvalidSignature):
        public_key2.verify(tbs, signature)


def test_register_unsupported_alg(generate_key):
    algorithms = [
        -18,  # SHAKE128, nonsensical value in this context
    ]
    with pytest.raises(CtapError) as exc_info:
        generate_key(algorithms)
    assert exc_info.value.code == CtapError.ERR.UNSUPPORTED_ALGORITHM


@pytest.mark.parametrize("flags", range(0, 256))
def test_register_invalid_flags(generate_key, flags):
    if flags in [0b000, 0b001, 0b101]:
        generate_key(ALL_ALGORITHMS, flags=flags)
    else:
        with pytest.raises(CtapError) as exc_info:
            generate_key(ALL_ALGORITHMS, flags=flags)

        assert exc_info.value.code == CtapError.ERR.INVALID_OPTION


def test_assert_empty_allow_list(ctap2, on_keepalive, credential_cache):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: (
            cred.flags == 0b000
            and (cred.options or {}).get("rk", False)
            and (cred.extensions or {}).get("credProtect", None) == 0x01
        ),
        ALL_ALGORITHMS,
        options={"rk": True},  # Required for allow_list=None to succees
        extensions={"credProtect": 0x01},  # Required for allow_list=None with up=False
    )
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)
    ext_inputs = sign_inputs(cred, tbs, additional_args=args)

    response1 = ctap2.get_assertion(
        RP["id"],
        EMPTY_JSON_HASH,
        allow_list=None,
        options={"up": False},
        extensions=None,
        on_keepalive=on_keepalive,
    )
    assert response1 is not None

    with pytest.raises(CtapError) as exc_info:
        ctap2.get_assertion(
            RP["id"],
            EMPTY_JSON_HASH,
            allow_list=None,
            options={"up": False},
            extensions={"previewSign": ext_inputs},
            on_keepalive=on_keepalive,
        )
    assert exc_info.value.code == CtapError.ERR.INVALID_OPTION


@pytest.mark.parametrize("missing_key", [AuthenticatorInput.KH, AuthenticatorInput.TBS])
def test_assert_missing_required_parameter(
    ctap2, on_keepalive, credential_cache, missing_key
):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.flags == 0b000, ALL_ALGORITHMS
    )
    credential_id = cred.response.auth_data.credential_data.credential_id
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)
    ext_inputs = sign_inputs(cred, tbs, additional_args=args)

    response1 = ctap2.get_assertion(
        RP["id"],
        EMPTY_JSON_HASH,
        allow_list=[{"type": "public-key", "id": credential_id}],
        options={"up": False},
        extensions={"previewSign": ext_inputs},
        on_keepalive=on_keepalive,
    )
    assert response1 is not None

    mod_ext_inputs = {**ext_inputs}
    del mod_ext_inputs[missing_key]
    with pytest.raises(CtapError) as exc_info:
        ctap2.get_assertion(
            RP["id"],
            EMPTY_JSON_HASH,
            allow_list=[{"type": "public-key", "id": credential_id}],
            options={"up": False},
            extensions={"previewSign": mod_ext_inputs},
            on_keepalive=on_keepalive,
        )
    assert exc_info.value.code == CtapError.ERR.INVALID_OPTION


def test_assert_missing_args_alg(ctap2, on_keepalive, credential_cache):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: (
            cred.algorithm == ESP256_SPLIT_ARKG_PLACEHOLDER and cred.flags == 0b000
        ),
        [ESP256_SPLIT_ARKG_PLACEHOLDER],
    )
    credential_id = cred.response.auth_data.credential_data.credential_id
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)
    if args is None:
        pytest.skip("Algorithm does not use additional arguments")
    ext_inputs = sign_inputs(cred, tbs, additional_args=args)

    response1 = ctap2.get_assertion(
        RP["id"],
        EMPTY_JSON_HASH,
        allow_list=[{"type": "public-key", "id": credential_id}],
        options={"up": False},
        extensions={"previewSign": ext_inputs},
        on_keepalive=on_keepalive,
    )
    assert response1 is not None

    del args[AuthenticatorInput.ALG]
    ext_inputs = sign_inputs(cred, tbs, additional_args=args)

    with pytest.raises(CtapError) as exc_info:
        ctap2.get_assertion(
            RP["id"],
            EMPTY_JSON_HASH,
            allow_list=[{"type": "public-key", "id": credential_id}],
            options={"up": False},
            extensions={"previewSign": ext_inputs},
            on_keepalive=on_keepalive,
        )
    assert exc_info.value.code == CtapError.ERR.INVALID_CREDENTIAL


def test_assert_incorrect_args_alg(ctap2, on_keepalive, credential_cache):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: (
            cred.algorithm == ESP256_SPLIT_ARKG_PLACEHOLDER and cred.flags == 0b000
        ),
        [ESP256_SPLIT_ARKG_PLACEHOLDER],
    )
    credential_id = cred.response.auth_data.credential_data.credential_id
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)
    if args is None:
        pytest.skip("Algorithm does not use additional arguments")
    ext_inputs = sign_inputs(cred, tbs, additional_args=args)

    response1 = ctap2.get_assertion(
        RP["id"],
        EMPTY_JSON_HASH,
        allow_list=[{"type": "public-key", "id": credential_id}],
        options={"up": False},
        extensions={"previewSign": ext_inputs},
        on_keepalive=on_keepalive,
    )
    assert response1 is not None

    args[AuthenticatorInput.ALG] += 1
    ext_inputs = sign_inputs(cred, tbs, additional_args=args)

    with pytest.raises(CtapError) as exc_info:
        ctap2.get_assertion(
            RP["id"],
            EMPTY_JSON_HASH,
            allow_list=[{"type": "public-key", "id": credential_id}],
            options={"up": False},
            extensions={"previewSign": ext_inputs},
            on_keepalive=on_keepalive,
        )
    assert exc_info.value.code == CtapError.ERR.INVALID_CREDENTIAL


def test_assert_missing_args(ctap2, on_keepalive, credential_cache, sign):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: (
            cred.algorithm == ESP256_SPLIT_ARKG_PLACEHOLDER and cred.flags == 0b000
        ),
        [ESP256_SPLIT_ARKG_PLACEHOLDER],
    )
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)
    if args is None:
        pytest.skip("Algorithm does not use additional arguments")
    sign_inputs(cred, tbs, additional_args=args)

    response1 = sign(cred, tbs, additional_args=args)
    assert response1 is not None

    with pytest.raises(CtapError) as exc_info:
        sign(cred, tbs, additional_args=None)
    assert exc_info.value.code == CtapError.ERR.MISSING_PARAMETER


def test_assert_unknown_args(ctap2, on_keepalive, credential_cache, sign):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: (
            cred.algorithm == ESP256_SPLIT_ARKG_PLACEHOLDER and cred.flags == 0b000
        ),
        [ESP256_SPLIT_ARKG_PLACEHOLDER],
    )
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)
    if args is None:
        pytest.skip("Algorithm does not use additional arguments")

    response1 = sign(cred, tbs, additional_args=args)
    assert response1 is not None

    response2 = sign(cred, tbs, additional_args={**args, 42: 1337})
    assert response2 is not None


def test_assert_up_required(ctap2, on_keepalive, credential_cache, sign):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.flags & 0b001 == 0b001,
        ALL_ALGORITHMS,
        flags=0b001,
    )
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)

    with pytest.raises(CtapError) as exc_info:
        sign(cred, tbs, additional_args=args, up=False)
    assert exc_info.value.code == CtapError.ERR.UP_REQUIRED


def test_assert_uv_required(ctap2, on_keepalive, credential_cache, sign):
    cred = credential_cache.make_cred_or_skip(
        lambda cred: cred.flags == 0b101,
        ALL_ALGORITHMS,
        flags=0b101,
    )
    tbs = os.urandom(32)
    public_key, args = if_arkg(cred.algorithm, cred.public_key)

    with pytest.raises(CtapError) as exc_info:
        sign(cred, tbs, additional_args=args, up=True)
    assert exc_info.value.code == CtapError.ERR.PUAT_REQUIRED



import pytest

from fido2.client import ClientError, DefaultClientDataCollector, Fido2Client
from fido2.ctap import CtapError
from fido2.ctap2.bio import BioEnrollment, CaptureError, FPBioEnrollment
from fido2.ctap2.pin import ClientPin
from fido2.server import Fido2Server

from . import TEST_PIN, CliInteraction


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if not BioEnrollment.is_supported(dev_manager.info):
        pytest.skip("BioEnrollment not supported by authenticator")
    if dev_manager.info.options["uv"]:
        pytest.skip("UV already configured")


def get_bio(ctap2, pin_protocol=None, permissions=ClientPin.PERMISSION.BIO_ENROLL):
    if pin_protocol:
        token = ClientPin(ctap2, pin_protocol).get_pin_token(TEST_PIN, permissions)
    else:
        token = None
    return FPBioEnrollment(ctap2, pin_protocol, token)


def test_get_sensor_info(ctap2):
    bio = get_bio(ctap2)
    info = bio.get_fingerprint_sensor_info()
    assert info.get(2) in (1, None)
    assert info.get(3, 1) > 0
    assert info.get(8, 1) > 0


def test_enroll_use_delete(device, ctap2, pin_protocol, printer):
    bio = get_bio(ctap2, pin_protocol)
    assert len(bio.enumerate_enrollments()) == 0

    context = bio.enroll()
    template_id = None
    while template_id is None:
        printer.print("Press your fingerprint against the sensor now...")
        try:
            template_id = context.capture()
            printer.print(f"{context.remaining} more scans needed.")
        except CaptureError as e:
            printer.print(e)

    enrollments = bio.enumerate_enrollments()
    assert len(enrollments) == 1
    assert enrollments[template_id] in ("", None)

    # Test name/rename
    info = bio.get_fingerprint_sensor_info()
    fname = "Test 1"
    bio.set_name(template_id, fname)

    enrollments = bio.enumerate_enrollments()
    assert len(enrollments) == 1
    assert enrollments[template_id] == fname

    fname = "Test".ljust(info.get(8, 0), "!")
    bio.set_name(template_id, fname)
    enrollments = bio.enumerate_enrollments()
    assert len(enrollments) == 1
    assert enrollments[template_id] == fname

    # Create a credential using fingerprint
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}
    create_options, state = server.register_begin(user, user_verification="required")

    client = Fido2Client(
        device,
        client_data_collector=DefaultClientDataCollector("https://example.com"),
        user_interaction=CliInteraction(printer, "WrongPin"),
    )

    try:
        # Allow multiple attempts
        for _ in range(3):
            try:
                result = client.make_credential(create_options.public_key)
                break
            except ClientError as e:
                if e.cause.code == CtapError.ERR.UV_INVALID:
                    continue
                raise

        server.register_complete(state, result)

    finally:
        # Delete fingerprint
        bio = get_bio(ctap2, pin_protocol)
        bio.remove_enrollment(template_id)
        assert len(bio.enumerate_enrollments()) == 0



import pytest
from fido2.client import ClientError, DefaultClientDataCollector, Fido2Client
from fido2.ctap import CtapError
from fido2.ctap2.config import Config
from fido2.ctap2.pin import ClientPin
from fido2.server import Fido2Server

from . import TEST_PIN, CliInteraction


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if not Config.is_supported(dev_manager.info):
        pytest.skip("Config not supported by authenticator")


@pytest.fixture
def client_pin(ctap2, pin_protocol):
    return ClientPin(ctap2, pin_protocol)


def get_config(
    ctap2,
    pin_protocol,
    pin=TEST_PIN,
    permissions=ClientPin.PERMISSION.AUTHENTICATOR_CFG,
):
    token = ClientPin(ctap2, pin_protocol).get_pin_token(pin, permissions)
    return Config(ctap2, pin_protocol, token)


def test_always_uv(ctap2, pin_protocol, device, printer):
    always_uv = ctap2.info.options.get("alwaysUv")
    if always_uv is None:
        pytest.skip("AlwaysUv not supported")

    if ctap2.info.options.get("uv"):
        pytest.skip("UV already configured")

    # Toggle on, if off
    if not always_uv:
        config = get_config(ctap2, pin_protocol)
        config.toggle_always_uv()

    assert ctap2.get_info().options["alwaysUv"] is True

    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    create_options, state = server.register_begin(user, user_verification="discouraged")

    # Create a credential
    client = Fido2Client(
        device,
        client_data_collector=DefaultClientDataCollector("https://example.com"),
        user_interaction=CliInteraction(printer, "WrongPin"),
    )

    # Should require PIN due to alwaysUV and fail
    with pytest.raises(ClientError, match="PIN_INVALID"):
        client.make_credential(create_options.public_key)

    # Toggle back off, if toggled on
    if not always_uv:
        config = get_config(ctap2, pin_protocol)
        config.toggle_always_uv()
        assert ctap2.get_info().options["alwaysUv"] is False

        # Now create the credential without requiring auth
        client.make_credential(create_options.public_key)


def test_force_pin_change(ctap2, pin_protocol, client_pin):
    assert ctap2.get_info().force_pin_change is False
    client_pin.get_pin_token(TEST_PIN)

    config = get_config(ctap2, pin_protocol)
    config.set_min_pin_length(force_change_pin=True)
    assert ctap2.get_info().force_pin_change is True

    with pytest.raises(CtapError, match="PIN_INVALID"):
        client_pin.get_pin_token(TEST_PIN)

    pin = TEST_PIN[::-1]
    client_pin.change_pin(TEST_PIN, pin)
    client_pin.change_pin(pin, TEST_PIN)
    client_pin.get_pin_token(TEST_PIN)


def test_min_pin_length(
    dev_manager, ctap2, pin_protocol, client_pin, printer, factory_reset
):
    config = get_config(ctap2, pin_protocol)

    orig_len = ctap2.info.min_pin_length
    expected_len = orig_len
    max_len = ctap2.info.max_pin_length
    if orig_len >= max_len:
        pytest.skip("Cannot increase min PIN length further")

    try:
        expected_len += 1
        config.set_min_pin_length(min_pin_length=expected_len)

        pin = TEST_PIN * 4

        # Too short
        with pytest.raises(CtapError, match="PIN_POLICY_VIOLATION"):
            client_pin.change_pin(TEST_PIN, pin[:orig_len])

        # Just long enough
        new_pin = pin[:expected_len]
        client_pin.change_pin(TEST_PIN, new_pin)

        if max_len >= orig_len + 2:
            # Even longer
            client_pin.change_pin(new_pin, pin[: expected_len + 1])
            # Change back
            client_pin.change_pin(pin[: expected_len + 1], new_pin)

        config = get_config(ctap2, pin_protocol, pin=new_pin)

        # Cannot shorten min pin length
        with pytest.raises(CtapError, match="PIN_POLICY_VIOLATION"):
            config.set_min_pin_length(min_pin_length=orig_len)

        if max_len >= orig_len + 2:
            expected_len = orig_len + 2
            config.set_min_pin_length(min_pin_length=expected_len)

            # Current PIN is too short
            assert ctap2.get_info().force_pin_change is True

            client_pin.change_pin(new_pin, pin[:expected_len])
            new_pin = pin[:expected_len]
            assert ctap2.get_info().force_pin_change is False

        # Test minPinLength extension
        rp = {"id": "example.com", "name": "Example RP"}
        server = Fido2Server(rp)
        user = {"id": b"user_id", "name": "A. User"}

        create_options, state = server.register_begin(
            user, user_verification="discouraged"
        )

        if "setMinPINLength" in ctap2.info.options:
            config = get_config(ctap2, pin_protocol, pin=new_pin)
            config.set_min_pin_length(rp_ids=[rp["id"]])
            client = Fido2Client(
                dev_manager.device,
                client_data_collector=DefaultClientDataCollector("https://example.com"),
                user_interaction=CliInteraction(printer, new_pin),
            )

            result = client.make_credential(
                {
                    **create_options["publicKey"],
                    "extensions": {"minPinLength": True},
                }
            )
            auth_data = server.register_complete(state, result)
            assert auth_data.extensions
            assert auth_data.extensions["minPinLength"] == expected_len

        if max_len > expected_len:
            # Increase min pin length to max
            config = get_config(ctap2, pin_protocol, pin=new_pin)
            config.set_min_pin_length(min_pin_length=max_len)

            assert ctap2.get_info().min_pin_length == max_len

            # Current PIN is too short
            assert ctap2.get_info().force_pin_change is True
    finally:
        # Restore original config
        factory_reset(setup=True)
        assert dev_manager.info.min_pin_length == orig_len


@pytest.fixture(scope="module")
def enable_ep(dev_manager, factory_reset):
    if "ep" not in dev_manager.info.options:
        pytest.skip("Enterprise Attestation not supported")

    assert dev_manager.info.options["ep"] is False

    # Enable EP
    pin_protocol = ClientPin(dev_manager.ctap2).protocol
    config = get_config(dev_manager.ctap2, pin_protocol)
    config.enable_enterprise_attestation()
    assert dev_manager.info.options["ep"] is True

    yield None

    # Restore original config
    factory_reset(setup=True)
    assert dev_manager.info.options["ep"] is False


@pytest.fixture(scope="module")
def att_cert(dev_manager):
    rp = {"id": "example.com", "name": "Example RP"}
    user = {"id": b"user_id", "name": "A. User"}

    server = Fido2Server(rp, attestation="direct")
    create_options, state = server.register_begin(user)
    result = dev_manager.client.make_credential(create_options.public_key)
    return result.response.attestation_object.att_stmt["x5c"][0]


def test_ep_platform(client, enable_ep, att_cert):
    rp = {"id": "example.com", "name": "Example RP"}
    user = {"id": b"user_id", "name": "A. User"}

    server = Fido2Server(rp, attestation="enterprise")
    create_options, state = server.register_begin(user)

    client._enterprise_rpid_list = [rp["id"]]
    result = client.make_credential(create_options.public_key)
    cert = result.response.attestation_object.att_stmt["x5c"][0]

    assert att_cert != cert


def test_ep_vendor(pytestconfig, device, printer, enable_ep, att_cert):
    ep_rp_id = pytestconfig.getoption("ep_rp_id")
    if not ep_rp_id:
        pytest.skip("No RP ID provided with --ep-rp-id")

    rp = {"id": ep_rp_id, "name": "Example RP"}
    user = {"id": b"user_id", "name": "A. User"}

    server = Fido2Server(rp, attestation="enterprise")
    create_options, state = server.register_begin(user)

    client = Fido2Client(
        device,
        client_data_collector=DefaultClientDataCollector(f"https://{ep_rp_id}"),
        user_interaction=CliInteraction(printer),
    )

    result = client.make_credential(create_options.public_key)
    cert = result.response.attestation_object.att_stmt["x5c"][0]

    assert att_cert != cert



# Copyright (c) 2025 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# flake8: noqa ignore lines too long


import pytest

from fido2.cose import ARKG_P256_PLACEHOLDER


TEST_VECTORS = [
    {
        "ctx": b"ARKG-P256.test vectors",
        "ikm_bl": "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f",
        "ikm_kem": "202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f",
        "ikm": "404142434445464748494a4b4c4d4e4f505152535455565758595a5b5c5d5e5f",
        "pk_bl": "046d3bdf31d0db48988f16d47048fdd24123cd286e42d0512daa9f726b4ecf18df65ed42169c69675f936ff7de5f9bd93adbc8ea73036b16e8d90adbfabdaddba7",
        "pk_kem": "04c38bbdd7286196733fa177e43b73cfd3d6d72cd11cc0bb2c9236cf85a42dcff5dfa339c1e07dfcdfda8d7be2a5a3c7382991f387dfe332b1dd8da6e0622cfb35",
        "sk_bl": "d959500a78ccf850ce46c80a8c5043c9a2e33844232b3829df37d05b3069f455",
        "sk_kem": "74e0a4cd81ca2d24246ff75bfd6d4fb7f9dfc938372627feb2c2348f8b1493b5",
        "pk_prime": "04572a111ce5cfd2a67d56a0f7c684184b16ccd212490dc9c5b579df749647d107dac2a1b197cc10d2376559ad6df6bc107318d5cfb90def9f4a1f5347e086c2cd",
        "sk_prime": "775d7fe9a6dfba43ce671cb38afca3d272c4d14aff97bd67559eb500a092e5e7",
    },
    {
        "ctx": b"ARKG-P256.test vectors",
        "ikm_bl": "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f",
        "ikm_kem": "202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f",
        "ikm": "a0a1a2a3a4a5a6a7a8a9aaabacadaeafb0b1b2b3b4b5b6b7b8b9babbbcbdbebf",
        "pk_bl": "046d3bdf31d0db48988f16d47048fdd24123cd286e42d0512daa9f726b4ecf18df65ed42169c69675f936ff7de5f9bd93adbc8ea73036b16e8d90adbfabdaddba7",
        "pk_kem": "04c38bbdd7286196733fa177e43b73cfd3d6d72cd11cc0bb2c9236cf85a42dcff5dfa339c1e07dfcdfda8d7be2a5a3c7382991f387dfe332b1dd8da6e0622cfb35",
        "sk_bl": "d959500a78ccf850ce46c80a8c5043c9a2e33844232b3829df37d05b3069f455",
        "sk_kem": "74e0a4cd81ca2d24246ff75bfd6d4fb7f9dfc938372627feb2c2348f8b1493b5",
        "pk_prime": "04ea7d962c9f44ffe8b18f1058a471f394ef81b674948eefc1865b5c021cf858f577f9632b84220e4a1444a20b9430b86731c37e4dcb285eda38d76bf758918d86",
        "sk_prime": "6228e470290e9d7cc0feff32a74caafa14c608c956337eba23997f5904cff226",
    },
    {
        "ctx": b"ARKG-P256.test vectors.0",
        "ikm_bl": "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f",
        "ikm_kem": "202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f",
        "ikm": "404142434445464748494a4b4c4d4e4f505152535455565758595a5b5c5d5e5f",
        "pk_bl": "046d3bdf31d0db48988f16d47048fdd24123cd286e42d0512daa9f726b4ecf18df65ed42169c69675f936ff7de5f9bd93adbc8ea73036b16e8d90adbfabdaddba7",
        "pk_kem": "04c38bbdd7286196733fa177e43b73cfd3d6d72cd11cc0bb2c9236cf85a42dcff5dfa339c1e07dfcdfda8d7be2a5a3c7382991f387dfe332b1dd8da6e0622cfb35",
        "sk_bl": "d959500a78ccf850ce46c80a8c5043c9a2e33844232b3829df37d05b3069f455",
        "sk_kem": "74e0a4cd81ca2d24246ff75bfd6d4fb7f9dfc938372627feb2c2348f8b1493b5",
        "pk_prime": "04b79b65d6bbb419ff97006a1bd52e3f4ad53042173992423e06e52987a037cb61dd82b126b162e4e7e8dc5c9fd86e82769d402a1968c7c547ef53ae4f96e10b0e",
        "sk_prime": "2a97f4232f9abba32fbfc28c6686f8afd2d851c2a95a3ed2f0a384b9ad55068d",
    },
]
# Convert hex strings to bytes
TEST_VECTORS = [
    {k: bytes.fromhex(v) if k != "ctx" else v for k, v in tv.items()}
    for tv in TEST_VECTORS
]


@pytest.mark.parametrize("test_vector", TEST_VECTORS)
def test_vectors(test_vector):
    pub_key = ARKG_P256_PLACEHOLDER(
        {
            1: -65537,
            3: -65700,
            -1: {
                1: 2,
                3: -7,
                -1: 1,
                -2: test_vector["pk_bl"][1:33],
                -3: test_vector["pk_bl"][33:65],
            },
            -2: {
                1: 2,
                3: -25,
                -1: 1,
                -2: test_vector["pk_kem"][1:33],
                -3: test_vector["pk_kem"][33:65],
            },
            -3: -9,
        }
    )

    pk_derived, args = pub_key.derive_public_key(test_vector["ikm"], test_vector["ctx"])
    point = b"\4" + pk_derived[-2] + pk_derived[-3]
    assert point.hex() == test_vector["pk_prime"].hex()



# Copyright (c) 2019 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from unittest import mock

import pytest
from fido2.hid import CTAPHID
from fido2.pcsc import CtapPcscDevice

CtapPcscDevice: type


@pytest.fixture(autouse=True, scope="module")
def preconditions():
    global CtapPcscDevice
    try:
        from fido2.pcsc import CtapPcscDevice
    except ImportError:
        pytest.skip("pyscard is not installed")


def test_pcsc_call_cbor():
    connection = mock.Mock()
    connection.transmit.side_effect = [(b"U2F_V2", 0x90, 0x00), (b"", 0x90, 0x00)]

    CtapPcscDevice(connection, "Mock")

    connection.transmit.assert_called_with(
        [0x80, 0x10, 0x80, 0x00, 0x01, 0x04, 0x00], None
    )


def test_pcsc_call_u2f():
    connection = mock.Mock()
    connection.transmit.side_effect = [
        (b"U2F_V2", 0x90, 0x00),
        (b"", 0x90, 0x00),
        (b"u2f_resp", 0x90, 0x00),
    ]

    dev = CtapPcscDevice(connection, "Mock")
    res = dev.call(CTAPHID.MSG, b"\x00\x01\x00\x00\x05" + b"\x01" * 5 + b"\x00")

    connection.transmit.assert_called_with(
        [0x00, 0x01, 0x00, 0x00, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00], None
    )
    assert res == b"u2f_resp\x90\x00"


def test_pcsc_call_version_2():
    connection = mock.Mock()
    connection.transmit.side_effect = [(b"U2F_V2", 0x90, 0x00), (b"", 0x90, 0x00)]

    dev = CtapPcscDevice(connection, "Mock")

    assert dev.version == 2


def test_pcsc_call_version_1():
    connection = mock.Mock()
    connection.transmit.side_effect = [(b"U2F_V2", 0x90, 0x00), (b"", 0x63, 0x85)]

    dev = CtapPcscDevice(connection, "Mock")

    assert dev.version == 1



import pytest

from fido2.ctap import CtapError
from fido2.ctap2.pin import ClientPin

from . import TEST_PIN


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if not ClientPin.is_supported(dev_manager.info):
        pytest.skip("ClientPin not supported by authenticator")


@pytest.fixture
def client_pin(ctap2, pin_protocol):
    return ClientPin(ctap2, pin_protocol)


def test_pin_validation(dev_manager, client_pin):
    assert dev_manager.ctap2.get_info().options["clientPin"] is True
    assert client_pin.get_pin_retries()[0] == 8

    # Wrong PIN decreases the retries remaining
    for retries in range(7, 4, -1):
        # Third attempt uses AUTH_BLOCKED
        with pytest.raises(CtapError, match="PIN_(INVALID|AUTH_BLOCKED)"):
            client_pin.get_pin_token("123456")
        assert client_pin.get_pin_retries()[0] == retries

    # Now soft-locked, does not decrement or unlock with any PIN
    for pin in (TEST_PIN, "123456"):
        with pytest.raises(CtapError, match="PIN_AUTH_BLOCKED"):
            client_pin.get_pin_token(pin)
    assert client_pin.get_pin_retries()[0] == retries

    dev_manager.reconnect()
    client_pin = ClientPin(dev_manager.ctap2, client_pin.protocol)

    # Wrong PIN decreases the retries remaining again
    with pytest.raises(CtapError, match="PIN_INVALID"):
        client_pin.get_pin_token("123456")
    assert client_pin.get_pin_retries()[0] == retries - 1

    # Unlocks with correct PIN
    token = client_pin.get_pin_token(TEST_PIN)
    assert client_pin.get_pin_retries()[0] == 8
    assert token


def test_change_pin(client_pin):
    client_pin.get_pin_token(TEST_PIN)

    new_pin = TEST_PIN[::-1]

    client_pin.change_pin(TEST_PIN, new_pin)
    with pytest.raises(CtapError, match="PIN_INVALID"):
        client_pin.get_pin_token(TEST_PIN)

    client_pin.get_pin_token(new_pin)

    client_pin.change_pin(new_pin, TEST_PIN)
    client_pin.get_pin_token(TEST_PIN)


def test_set_and_reset(dev_manager, client_pin, factory_reset):
    assert dev_manager.ctap2.get_info().options["clientPin"] is True
    assert client_pin.get_pin_retries()[0] == 8

    factory_reset()
    client_pin = ClientPin(dev_manager.ctap2, client_pin.protocol)
    # Factory reset clears the PIN
    assert dev_manager.ctap2.get_info().options["clientPin"] is False
    with pytest.raises(CtapError, match="PIN_NOT_SET"):
        client_pin.get_pin_retries()

    # Setup includes setting the default PIN. More correct would be to just set
    # the PIN ourselves here and test that, but then we need another factory reset
    dev_manager.setup()
    assert dev_manager.ctap2.get_info().options["clientPin"] is True
    assert client_pin.get_pin_retries()[0] == 8



from fido2.server import Fido2Server


def test_make_assert(client, pin_protocol, algorithm):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    create_options, state = server.register_begin(user)

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "pubKeyCredParams": [algorithm],
        }
    )

    auth_data = server.register_complete(state, result)
    cred = auth_data.credential_data
    assert cred.public_key[3] == algorithm["alg"]
    credentials = [cred]

    # Get assertion
    request_options, state = server.authenticate_begin(credentials)

    # Authenticate the credential
    result = client.get_assertion(request_options.public_key).get_response(0)
    cred_data = server.authenticate_complete(state, credentials, result)
    assert cred_data == cred



from fido2.client import UserInteraction

TEST_PIN = "a1b2c3d4"


class Printer:
    def __init__(self, capmanager):
        self.capmanager = capmanager

    def print(self, *messages):
        with self.capmanager.global_and_fixture_disabled():
            print("")
            for m in messages:
                print(m)

    def touch(self):
        self.print("👉 Touch the Authenticator")

    def insert(self, nfc=False):
        self.print(
            "♻️  "
            + (
                "Place the Authenticator on the NFC reader"
                if nfc
                else "Connect the Authenticator"
            )
        )

    def remove(self, nfc=False):
        self.print(
            "🚫 "
            + (
                "Remove the Authenticator from the NFC reader"
                if nfc
                else "Disconnect the Authenticator"
            )
        )


# Handle user interaction
class CliInteraction(UserInteraction):
    def __init__(self, printer, pin=TEST_PIN):
        self.printer = printer
        self.pin = pin

    def prompt_up(self):
        self.printer.touch()

    def request_pin(self, permissions, rp_id):
        return self.pin

    def request_uv(self, permissions, rp_id):
        self.printer.print("User Verification required.")
        return True



from fido2.webauthn import Aaguid


def assert_list_of(typ, value):
    assert isinstance(value, list)
    for v in value:
        assert isinstance(v, typ)


def assert_dict_of(k_type, v_type, value):
    assert isinstance(value, dict)
    for k, v in value.items():
        assert isinstance(k, k_type)
        assert isinstance(v, v_type)


def assert_unique(value):
    assert len(set(value)) == len(value)


def test_get_info_fields(ctap2):
    info = ctap2.get_info()

    assert_list_of(str, info.versions)
    assert len(info.versions) > 0

    assert_list_of(str, info.extensions)
    assert isinstance(info.aaguid, Aaguid)
    assert_dict_of(str, bool | None, info.options)
    assert isinstance(info.max_msg_size, int)
    assert_list_of(int, info.pin_uv_protocols)
    assert_unique(info.pin_uv_protocols)
    assert isinstance(info.max_creds_in_list, int)
    assert isinstance(info.max_cred_id_length, int)
    assert_list_of(str, info.transports)
    assert_unique(info.transports)

    assert_list_of(dict, info.algorithms)
    assert isinstance(info.max_large_blob, int)
    assert isinstance(info.force_pin_change, bool)
    assert isinstance(info.min_pin_length, int)
    assert info.min_pin_length >= 4
    assert isinstance(info.firmware_version, int)
    assert isinstance(info.max_cred_blob_length, int)
    assert isinstance(info.max_rpids_for_min_pin, int)
    assert isinstance(info.preferred_platform_uv_attempts, int)
    assert isinstance(info.uv_modality, int)
    assert_dict_of(str, int, info.certifications)

    assert isinstance(info.remaining_disc_creds, int | None)
    assert_list_of(int, info.vendor_prototype_config_commands)
    assert_list_of(str, info.attestation_formats)
    assert_unique(info.attestation_formats)
    assert len(info.attestation_formats) > 0

    assert isinstance(info.uv_count_since_pin, int | None)
    assert isinstance(info.long_touch_for_reset, bool)


def test_enc_identifier_changes(ctap2):
    if ctap2.info.enc_identifier:
        assert ctap2.get_info().enc_identifier != ctap2.get_info().enc_identifier



import os

import pytest

from fido2.client import DefaultClientDataCollector, Fido2Client
from fido2.ctap2.extensions import HmacSecretExtension
from fido2.server import Fido2Server
from fido2.utils import websafe_encode

from . import TEST_PIN, CliInteraction


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if "hmac-secret" not in dev_manager.info.extensions:
        pytest.skip("hmac-secret not supported by authenticator")


def test_prf(client, pin_protocol):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}
    uv = "required"

    create_options, state = server.register_begin(user, user_verification=uv)

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"prf": {}},
        }
    )
    assert result.client_extension_results.prf.enabled is True
    assert result.client_extension_results["prf"]["enabled"] is True

    auth_data = server.register_complete(state, result)
    credentials = [auth_data.credential_data]

    # Complete registration
    auth_data = server.register_complete(state, result)
    credential = auth_data.credential_data

    # Generate a salt for PRF:
    salt = websafe_encode(os.urandom(32))

    # Prepare parameters for getAssertion
    credentials = [credential]
    request_options, state = server.authenticate_begin(
        credentials, user_verification=uv
    )

    # Authenticate the credential
    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {"prf": {"eval": {"first": salt}}},
        }
    )

    # Only one cred in allowCredentials, only one response.
    response = result.get_response(0)

    output1 = response.client_extension_results.prf.results.first
    assert response.client_extension_results["prf"]["results"][
        "first"
    ] == websafe_encode(output1)

    # Authenticate again, using two salts to generate two secrets.

    # This time we will use evalByCredential, which can be used if there are multiple
    # credentials which use different salts. Here it is not needed, but provided for
    # completeness of the example.

    # Generate a second salt for PRF:
    salt2 = websafe_encode(os.urandom(32))
    # The first salt is reused, which should result in the same secret.

    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {
                "prf": {
                    "evalByCredential": {
                        websafe_encode(credential.credential_id): {
                            "first": salt,
                            "second": salt2,
                        }
                    }
                }
            },
        }
    )

    response = result.get_response(0)

    output = response.client_extension_results.prf.results
    assert output.first == output1
    assert output.second != output1
    assert response.client_extension_results["prf"]["results"][
        "second"
    ] == websafe_encode(output.second)


def test_hmac_secret(device, pin_protocol, printer):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}
    uv = "required"

    create_options, state = server.register_begin(user, user_verification=uv)

    client = Fido2Client(
        device,
        client_data_collector=DefaultClientDataCollector("https://example.com"),
        user_interaction=CliInteraction(printer, TEST_PIN),
        extensions=[HmacSecretExtension(allow_hmac_secret=True)],
    )

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"hmacCreateSecret": True},
        }
    )
    assert result.client_extension_results.hmac_create_secret is True
    assert result.client_extension_results["hmacCreateSecret"] is True

    # Complete registration
    auth_data = server.register_complete(state, result)
    credentials = [auth_data.credential_data]

    # Generate a salt for HmacSecret:
    salt = os.urandom(32)

    # Prepare parameters for getAssertion
    request_options, state = server.authenticate_begin(
        credentials, user_verification=uv
    )

    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {"hmacGetSecret": {"salt1": salt}},
        }
    )
    result = result.get_response(0)

    output1 = result.client_extension_results.hmac_get_secret.output1
    assert result.client_extension_results["hmacGetSecret"][
        "output1"
    ] == websafe_encode(output1)

    salt2 = os.urandom(32)

    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {"hmacGetSecret": {"salt1": salt, "salt2": salt2}},
        }
    )
    result = result.get_response(0)

    output = result.client_extension_results.hmac_get_secret
    assert output.output1 == output1
    assert output.output2 != output1


def test_prf_mc(client, pin_protocol, info):
    if "hmac-secret-mc" not in info.extensions:
        pytest.skip("hmac-secret-mc not supported by authenticator")

    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}
    uv = "required"

    create_options, state = server.register_begin(user, user_verification=uv)

    # Generate salts for PRF:
    salt1 = websafe_encode(os.urandom(32))
    salt2 = websafe_encode(os.urandom(32))

    # Create a credential, with salts
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"prf": {"eval": {"first": salt1, "second": salt2}}},
        }
    )
    auth_data = server.register_complete(state, result)
    credential = auth_data.credential_data

    assert result.client_extension_results.prf.enabled is True
    assert result.client_extension_results["prf"]["enabled"] is True

    output = result.client_extension_results.prf.results
    assert output.first
    assert output.second

    # Prepare parameters for getAssertion
    credentials = [credential]
    request_options, state = server.authenticate_begin(
        credentials, user_verification=uv
    )

    # Authenticate the credential
    result = client.get_assertion(
        {
            **request_options["publicKey"],
            "extensions": {
                "prf": {
                    "evalByCredential": {
                        websafe_encode(credential.credential_id): {
                            "first": salt1,
                            "second": salt2,
                        }
                    }
                }
            },
        }
    )

    response = result.get_response(0)
    assert output == response.client_extension_results.prf.results



import pytest
from fido2.hid import CtapHidDevice


def test_ping(device):
    if not isinstance(device, CtapHidDevice):
        pytest.skip("Device is not a CtapHidDevice")

    msg1 = b"hello world!"
    msg2 = b"            "
    msg3 = b""
    assert device.ping(msg1) == msg1
    assert device.ping(msg2) == msg2
    assert device.ping(msg3) == msg3



# Copyright (c) 2018 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Connects to the first FIDO device found which supports the HmacSecret extension,
creates a new credential for it with the extension enabled, and uses it to
derive two separate secrets.

NOTE: This extension is not enabled by default as direct access to the extension
is now allowed in a browser setting. See also prf.py for an example which uses
the PRF extension which is enabled by default.
"""

import ctypes
import os
import sys

from exampleutils import CliInteraction, enumerate_devices

from fido2.client import DefaultClientDataCollector, Fido2Client
from fido2.ctap2.extensions import HmacSecretExtension
from fido2.server import Fido2Server

# Use the Windows WebAuthn API if available, and we're not running as admin
try:
    from fido2.client.windows import WindowsClient

    use_winclient = (
        WindowsClient.is_available() and not ctypes.windll.shell32.IsUserAnAdmin()
    )
except ImportError:
    use_winclient = False


uv = "discouraged"

client_data_collector = DefaultClientDataCollector("https://example.com")

if use_winclient:
    # Use the Windows WebAuthn API if available, and we're not running as admin
    # By default only the PRF extension is allowed, we need to explicitly
    # configure the client to allow hmac-secret
    client = WindowsClient(client_data_collector, allow_hmac_secret=True)
else:
    # Locate a device
    for dev in enumerate_devices():
        client = Fido2Client(
            dev,
            client_data_collector=client_data_collector,
            user_interaction=CliInteraction(),
            # By default only the PRF extension is allowed, we need to explicitly
            # configure the client to allow hmac-secret
            extensions=[HmacSecretExtension(allow_hmac_secret=True)],
        )
        if "hmac-secret" in client.info.extensions:
            break
    else:
        print("No Authenticator with the HmacSecret extension found!")
        sys.exit(1)

server = Fido2Server({"id": "example.com", "name": "Example RP"}, attestation="none")
user = {"id": b"user_id", "name": "A. User"}

# Prepare parameters for makeCredential
create_options, state = server.register_begin(
    user,
    resident_key_requirement="discouraged",
    user_verification=uv,
    authenticator_attachment="cross-platform",
)

# Create a credential
result = client.make_credential(
    {
        **create_options["publicKey"],
        "extensions": {"hmacCreateSecret": True},
    }
)

# Complete registration
auth_data = server.register_complete(state, result)
credentials = [auth_data.credential_data]

# HmacSecret result:
if result.client_extension_results.get("hmacCreateSecret"):
    print("New credential created, with HmacSecret")
else:
    # This fails on Windows, but we might still be able to use hmac-secret even if
    # the credential wasn't made with it, so keep going
    print("Failed to create credential with HmacSecret, it might not work")

# Generate a salt for HmacSecret:
salt = os.urandom(32)
print("Authenticate with salt:", salt.hex())


# Prepare parameters for getAssertion
request_options, state = server.authenticate_begin(credentials, user_verification=uv)

# Authenticate the credential
result = client.get_assertion(
    {
        **request_options["publicKey"],
        "extensions": {"hmacGetSecret": {"salt1": salt}},
    }
)

# Only one cred in allowCredentials, only one response.
result = result.get_response(0)

output1 = result.client_extension_results.hmac_get_secret.output1
print("Authenticated, secret:", output1.hex())

# Authenticate again, using two salts to generate two secrets:

# Generate a second salt for HmacSecret:
salt2 = os.urandom(32)
print("Authenticate with second salt:", salt2.hex())

# The first salt is reused, which should result in the same secret.

result = client.get_assertion(
    {
        **request_options["publicKey"],
        "extensions": {"hmacGetSecret": {"salt1": salt, "salt2": salt2}},
    }
)

# Only one cred in allowCredentials, only one response.
result = result.get_response(0)

output = result.client_extension_results.hmac_get_secret
print("Old secret:", output.output1.hex())
print("New secret:", output.output2.hex())



import os

import pytest

from fido2.client import ClientError
from fido2.ctap2.credman import CredentialManagement
from fido2.ctap2.pin import ClientPin
from fido2.server import Fido2Server, to_descriptor

from . import TEST_PIN

rp = {"id": "example.com", "name": "Example RP"}
user = {"id": b"user_id", "name": "A. User"}
server = Fido2Server(rp)


@pytest.fixture(scope="module")
def excluded_match(dev_manager):
    if dev_manager.has_ctap2():
        return "CREDENTIAL_EXCLUDED"
    return "DEVICE_INELIGIBLE"


@pytest.fixture(scope="module")
def credential(dev_manager):
    create_options, state = server.register_begin(user)
    result = dev_manager.client.make_credential(create_options.public_key)
    auth_data = server.register_complete(state, result)
    return auth_data.credential_data


@pytest.fixture(scope="module")
def discoverable_credential(request, dev_manager):
    if not dev_manager.has_ctap2():
        pytest.skip("Authenticator does not support CTAP 2")

    has_credman = CredentialManagement.is_supported(dev_manager.info)
    if not has_credman:
        # Request dynamically since we don't want to skip the test unless needed
        factory_reset = request.getfixturevalue("factory_reset")

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
    )
    result = dev_manager.client.make_credential(create_options.public_key)
    auth_data = server.register_complete(state, result)
    yield auth_data.credential_data

    # Delete credential via credman, or factory reset
    if has_credman:
        client_pin = ClientPin(dev_manager.ctap2)
        token = client_pin.get_pin_token(TEST_PIN, ClientPin.PERMISSION.CREDENTIAL_MGMT)
        credman = CredentialManagement(dev_manager.ctap2, client_pin.protocol, token)
        cred_id = {"id": auth_data.credential_data.credential_id, "type": "public-key"}
        credman.delete_cred(cred_id)
    else:
        factory_reset(setup=True)


def test_exclude_credentials_single(credential, client, excluded_match):
    create_options, state = server.register_begin(user, [credential])
    with pytest.raises(ClientError, match=excluded_match):
        client.make_credential(create_options.public_key)


def test_exclude_credentials_multiple(credential, client, excluded_match):
    exclude = [{"id": os.urandom(32), "type": "public-key"} for _ in range(5)]
    exclude.insert(3, to_descriptor(credential))
    create_options, state = server.register_begin(user, exclude)
    with pytest.raises(ClientError, match=excluded_match):
        client.make_credential(create_options.public_key)


def test_exclude_credentials_max(info, client):
    max_len = info.max_cred_id_length
    n_creds = (info.max_creds_in_list or 1) + 2
    exclude = [
        {"id": os.urandom(max_len), "type": "public-key"} for _ in range(n_creds)
    ]
    create_options, state = server.register_begin(user, exclude)
    client.make_credential(create_options.public_key)


def test_exclude_credentials_others(client):
    exclude = [{"id": os.urandom(32), "type": "public-key"} for _ in range(5)]
    create_options, state = server.register_begin(user, exclude)
    client.make_credential(create_options.public_key)


def test_allow_credentials_empty(discoverable_credential, client):
    request_options, state = server.authenticate_begin()
    result = client.get_assertion(request_options.public_key).get_response(0)
    server.authenticate_complete(state, [discoverable_credential], result)


def test_allow_credentials_single(credential, client):
    credentials = [credential]
    request_options, state = server.authenticate_begin(credentials)
    result = client.get_assertion(request_options.public_key).get_response(0)
    server.authenticate_complete(state, credentials, result)


def test_allow_credentials_multiple(credential, client):
    credentials = [credential]
    allow = [{"id": os.urandom(32), "type": "public-key"} for _ in range(5)]
    allow.insert(3, to_descriptor(credential))
    request_options, state = server.authenticate_begin(allow)
    result = client.get_assertion(request_options.public_key).get_response(0)
    server.authenticate_complete(state, credentials, result)


def test_allow_credentials_ineligible(client):
    allow = [{"id": os.urandom(32), "type": "public-key"} for _ in range(5)]
    request_options, state = server.authenticate_begin(allow)
    with pytest.raises(ClientError, match="DEVICE_INELIGIBLE"):
        client.get_assertion(request_options.public_key)



import os
import struct

import pytest

from fido2 import cbor
from fido2.ctap import CtapError
from fido2.ctap2.blob import LargeBlobs
from fido2.ctap2.pin import ClientPin
from fido2.server import Fido2Server
from fido2.utils import sha256, websafe_decode, websafe_encode

from . import TEST_PIN


@pytest.fixture(autouse=True, scope="module")
def preconditions(dev_manager):
    if not LargeBlobs.is_supported(dev_manager.info):
        pytest.skip("LargeBlobs not supported by authenticator")


def get_lb(ctap2, pin_protocol, permissions=ClientPin.PERMISSION.LARGE_BLOB_WRITE):
    token = ClientPin(ctap2, pin_protocol).get_pin_token(TEST_PIN, permissions)
    return LargeBlobs(ctap2, pin_protocol, token)


def test_read_write(ctap2, pin_protocol):
    lb = get_lb(ctap2, pin_protocol)
    assert len(lb.read_blob_array()) == 0

    key1 = os.urandom(32)
    data1 = b"test data"
    key2 = os.urandom(32)
    data2 = b"some other data"

    assert lb.get_blob(key1) is None
    lb.put_blob(key1, data1)
    assert lb.get_blob(key1) == data1
    assert len(lb.read_blob_array()) == 1

    lb.put_blob(key2, data2)
    assert lb.get_blob(key1) == data1
    assert lb.get_blob(key2) == data2
    assert len(lb.read_blob_array()) == 2

    lb.delete_blob(key1)
    assert lb.get_blob(key1) is None
    assert lb.get_blob(key2) == data2
    assert len(lb.read_blob_array()) == 1

    lb.delete_blob(key2)
    assert lb.get_blob(key2) is None
    assert len(lb.read_blob_array()) == 0


def test_invalid_checksum(ctap2, pin_protocol):
    lb = get_lb(ctap2, pin_protocol)

    data = cbor.encode([])
    # Set the checksum to an invalid value to ensure the authenticator checks it
    data += sha256(data)[:15] + b"\xff"
    offset = 0
    size = len(data)

    msg = b"\xff" * 32 + b"\x0c\x00" + struct.pack("<I", offset) + sha256(data)
    pin_uv_protocol = pin_protocol.VERSION
    pin_uv_param = pin_protocol.authenticate(lb.pin_uv.token, msg)

    with pytest.raises(CtapError, match="INTEGRITY_FAILURE"):
        ctap2.large_blobs(
            0,
            set=data,
            length=size,
            pin_uv_protocol=pin_uv_protocol,
            pin_uv_param=pin_uv_param,
        )


def test_size_bounds(ctap2, pin_protocol):
    lb = get_lb(ctap2, pin_protocol)
    assert len(lb.read_blob_array()) == 0

    # The max storage is the CBOR-encoded array, minus a 16 byte checksum
    max_size = ctap2.info.max_large_blob - 16

    # Create data which when cbor-encoded is exactly max_size bytes
    array = [{1: os.urandom(max_size - 8)}]
    array.extend([0] * (max_size - len(cbor.encode(array))))

    lb.write_blob_array(array)

    # Ensure writing larger data fails:
    array.append(1)
    with pytest.raises(CtapError, match="LARGE_BLOB_STORAGE_FULL"):
        lb.write_blob_array(array)

    # Clear the data
    lb.write_blob_array([])


def test_missing_permissions(ctap2, pin_protocol):
    key = os.urandom(32)
    data = b"test data"

    # Try write without PIN token
    lb = LargeBlobs(ctap2, pin_protocol)
    blobs = lb.read_blob_array()
    assert len(blobs) == 0

    with pytest.raises(CtapError, match="PUAT_REQUIRED"):
        lb.put_blob(key, data)

    # Try with wrong permissions
    lb = get_lb(ctap2, pin_protocol, ClientPin.PERMISSION.CREDENTIAL_MGMT)
    with pytest.raises(CtapError, match="PIN_AUTH_INVALID"):
        lb.put_blob(key, data)


def test_large_blob_key(client, ctap2, pin_protocol, clear_creds):
    rp = {"id": "example.com", "name": "Example RP"}
    server = Fido2Server(rp)
    user = {"id": b"user_id", "name": "A. User"}

    create_options, state = server.register_begin(
        user,
        resident_key_requirement="required",
        user_verification="required",
    )

    # Create a credential
    result = client.make_credential(
        {
            **create_options["publicKey"],
            "extensions": {"largeBlob": {"support": "required"}},
        }
    )
    auth_data = server.register_complete(state, result)
    credentials = [auth_data.credential_data]

    assert result.client_extension_results.large_blob.supported is True
    assert result.client_extension_results["largeBlob"]["supported"] is True

    request_options, state = server.authenticate_begin(
        credentials, user_verification="required"
    )

    data = b"test data"

    selection = client.get_assertion(
        {
            **request_options["publicKey"],
            # Write a large blob
            "extensions": {"largeBlob": {"write": websafe_encode(data)}},
        }
    )
    result = selection.get_response(0)

    assert result.client_extension_results.large_blob.written is True
    assert result.client_extension_results["largeBlob"]["written"] is True

    # Authenticate the credential
    selection = client.get_assertion(
        {
            **request_options["publicKey"],
            # Read back the blob
            "extensions": {"largeBlob": {"read": True}},
        }
    )
    result = selection.get_response(0)

    assert result.client_extension_results.large_blob.blob == data
    assert websafe_decode(result.client_extension_results["largeBlob"]["blob"]) == data

    # Clear the data
    lb = get_lb(ctap2, pin_protocol)
    lb.write_blob_array([])
    assert len(lb.read_blob_array()) == 0
